#include <nvs_flash.h>
#include <cJSON.h>
#include <string>

constexpr char *nvs_wifi_creds_namespace{"wifi_creds"};
constexpr char *nvs_wifi_creds_key{"creds_key"};

int cred_key{0};

void readWiFiCredentials()
{
    esp_err_t err;
    nvs_handle wifi_creds_handle;

    // Opening namespace
    err = nvs_open(nvs_wifi_creds_namespace, NVS_READONLY, &wifi_creds_handle);

    switch (err)
    {
    case ESP_OK:
        ESP_LOGI(__func__, "file opened");
        break;
    case ESP_ERR_NVS_NOT_FOUND:
        ESP_LOGE(__func__, "NVS not found!");
        return;
        break;
    default:
        break;
    }

    // Retrieving credentials list from NVS
    size_t required_size;
    err = nvs_get_str(wifi_creds_handle, nvs_wifi_creds_key, NULL, &required_size);
    ESP_LOGD(__func__, "Required size is: %i", required_size);

    if (err == ESP_ERR_NVS_NOT_FOUND || required_size == 0)
    {
        ESP_LOGE(__func__, "NVS not found, or required size is 0");
        return;
    }

    char *creds_list_str = new char[required_size];
    err = nvs_get_str(wifi_creds_handle, nvs_wifi_creds_key, creds_list_str, &required_size);

    ESP_LOGI(__func__, "Credentials retrieved: \n%s", creds_list_str);

    // Freeing array
    delete[] creds_list_str;

    nvs_close(wifi_creds_handle);
}

//====================================================================================================

void updateWiFiCredentials(const char *ssid, const char *password)
{
    ESP_LOGD(__func__, "SSID to add: %s", ssid);
    ESP_LOGD(__func__, "Password to add: %s", password);

    nvs_handle wifi_creds_handle;
    esp_err_t err;

    // Build JSON data
    cJSON *creds_list_json = cJSON_CreateObject();
    cJSON *new_creds = cJSON_CreateObject();
    cJSON *creds_num = cJSON_CreateObject();

    // Storing new credentials comming on a JSON object
    cJSON *comming_creds = cJSON_CreateObject();
    cJSON_AddItemToObject(comming_creds, "ssid", cJSON_CreateString(ssid));
    cJSON_AddItemToObject(comming_creds, "password", cJSON_CreateString(password));

    // Opening namespace
    err = nvs_open(nvs_wifi_creds_namespace, NVS_READWRITE, &wifi_creds_handle);
    if (err != ESP_OK)
    {
        ESP_LOGE(__func__, "Error %i on nvs_open()", err);
        ESP_ERROR_CHECK(err);
    }

    // Retrieving existing credentials list from NVS
    size_t required_size;
    err = nvs_get_str(wifi_creds_handle, nvs_wifi_creds_key, NULL, &required_size);
    if (err == ESP_ERR_NVS_NOT_FOUND || required_size == 0)
    {
        ESP_LOGD(__func__, "NVS not found!");
        // Store the comming credentials JSON object with the initial index of the list
        cJSON_AddItemToObject(new_creds, String(cred_key).c_str(), comming_creds);
        ESP_LOGI(__func__, "JSON data: \n%s", cJSON_Print(new_creds));

        // Storing the new credentials list on the NVS
        ESP_LOGD(__func__, "Updating credentials list...");
        err = nvs_set_str(wifi_creds_handle, nvs_wifi_creds_key, cJSON_Print(new_creds));
        if (err != ESP_OK)
        {
            ESP_LOGE(__func__, "Error %i on nvs_set_str()", err);
            ESP_ERROR_CHECK(err);
        }
    }
    else
    {
        int creds_counter{0};
        bool replace = false;
        // Read the existing credentials list first
        char *creds_list_str = new char[required_size];
        err = nvs_get_str(wifi_creds_handle, nvs_wifi_creds_key, creds_list_str, &required_size);
        // Storing old credentials on a JSON object
        creds_list_json = cJSON_Parse(creds_list_str);

        // Read what's the next key number in the credentials list and if any of them matches with the credentials comming
        while (cJSON_GetObjectItem(creds_list_json, String(cred_key).c_str()) != nullptr)
        {
            // Verify if new credentials match with any credentials already stored
            creds_num = cJSON_GetObjectItem(creds_list_json, String(cred_key).c_str());
            //td::cout << '\n'<< "Credentials Item: " << creds_num.toString() << '\n';

            std::string ssid_str(cJSON_GetObjectItem(creds_num, "ssid")->valuestring);
            // Cleaning up the string ssid string obtained
            ssid_str.erase(std::remove(ssid_str.begin(), ssid_str.end(), '['), ssid_str.end());
            ssid_str.erase(std::remove(ssid_str.begin(), ssid_str.end(), ']'), ssid_str.end());
            ssid_str.erase(std::remove(ssid_str.begin(), ssid_str.end(), '\"'), ssid_str.end());
            //std::cout << '\n' << "SSID_STR: " << ssid_str << '\n';

            if (ssid_str == std::string(ssid))
            {
                ESP_LOGI(__func__, "Credentials already stored, so the infomation will be replaced");
                replace = true;
                break;
            }
            cred_key++;
        }

        // Storing the new credentials list on the NVS
        ESP_LOGD(__func__, "Updating credentials list...");

        if (replace)
        {
            // Storing the comming credentials to the highest number of priority
            //cJSON_ReplaceItemInObject(creds_list_json, String(cred_key).c_str(), comming_creds);

            // Determine how many credentials are already stored in the file
            int creds_counter{0}; // Counts how many credentials are stored in the system
            while (cJSON_GetObjectItem(creds_list_json, String(creds_counter).c_str()) != nullptr)
            {
                creds_counter++;
            }
            ESP_LOGD(__func__, "Credentials counter: %i", creds_counter);
            if (creds_counter > 1)
            {
                for (int i = 0, j = 0; i < creds_counter && j < creds_counter; i++, j++)
                {
                    // Retreiving all the old credentials, except the matched one, to be set as highest priority after
                    if (i != cred_key)
                    {
                        // cJSON_ReplaceItemInObject(creds_list_json, String(i).c_str(), cJSON_GetObjectItem(creds_list_json, String(j).c_str()));
                        cJSON_AddItemToObject(new_creds, String(i).c_str(), cJSON_GetObjectItem(creds_list_json, String(j).c_str()));
                    }
                    else
                    {
                        j++;
                        if (j != creds_counter)
                        {
                            // cJSON_ReplaceItemInObject(creds_list_json, String(i).c_str(), cJSON_GetObjectItem(creds_list_json, String(j).c_str()));
                            cJSON_AddItemToObject(new_creds, String(i).c_str(), cJSON_GetObjectItem(creds_list_json, String(j).c_str()));
                        }
                        else
                        {
                            break;
                        }
                    }
                }
            }

            // At this point the older or lower priority credentials are re-ordered

            cJSON_AddItemToObject(new_creds, String(creds_counter - 1).c_str(), comming_creds);

            // Storing the updated credentials list on NVS
            err = nvs_set_str(wifi_creds_handle, nvs_wifi_creds_key, cJSON_Print(new_creds));
            if (err != ESP_OK)
            {
                ESP_LOGE(__func__, "Error %i on nvs_set_str()", err);
                ESP_ERROR_CHECK(err);
            }

            // Storing the comming credentials to the highest number of priority
            // cJSON_ReplaceItemInObject(old_creds, std::to_string(creds_counter - 1).c_str(), comming_creds);
            // It's about the highest priority credential, just update it
            // cJSON_ReplaceItemInObject(creds_list_json, String(creds_counter - 1).c_str(), comming_creds);
            // cJSON_AddItemToObject(creds_list_json, String(creds_counter - 1).c_str(), comming_creds);
        }
        else
        {
            // Just add the comming credentials to the creds_list_json object
            cJSON_AddItemToObject(creds_list_json, String(cred_key).c_str(), comming_creds);
            // Storing the updated credentials list on NVS
            err = nvs_set_str(wifi_creds_handle, nvs_wifi_creds_key, cJSON_Print(creds_list_json));
            if (err != ESP_OK)
            {
                ESP_LOGE(__func__, "Error %i on nvs_set_str()", err);
                ESP_ERROR_CHECK(err);
            }
        }

        // Freeing array
        delete[] creds_list_str;
    }

    // Commiting changes
    nvs_commit(wifi_creds_handle);
    // Closing NVS namespace
    nvs_close(wifi_creds_handle);
}

void setup()
{
    Serial.begin(115200);
    ESP_ERROR_CHECK(nvs_flash_init());
    ESP_ERROR_CHECK(nvs_flash_erase());
    /*
  int creds_counter{2};
  int cred_key{1};

  if (creds_counter > 1) {
    for (int i = 0, j = 0; i < creds_counter && j < creds_counter; i++, j++) {
      // Retreiving all the old credentials, except the matched one, to be set as highest priority after
      if (i != cred_key) {
        Serial.println("Not cred_key");
        Serial.print("i = ");
        Serial.println(i);
        Serial.print("j = ");
        Serial.println(j);
      } else {
        Serial.println("Yes cred_key");
        //j += 1;
        j++;
        Serial.print("i = ");
        Serial.println(i);
        Serial.print("j = ");
        Serial.println(j);
        if (j != creds_counter) {
          Serial.println("j is NOT the highest index");
        } else {
          Serial.println("j is the highest index");
        }
      }
    }
  }
  */

    updateWiFiCredentials("random", "access1234");
    updateWiFiCredentials("HELLO", "takemypassword");
    //readWiFiCredentials();
    updateWiFiCredentials("HELLO", "donttakeitplease");
    readWiFiCredentials();
}

void loop()
{
}
